unit ExHeaderControl;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, CommCtrl, Registry, DBugIntf;

{*******************************************************************************
* Extended Header definitions from commctrl.h                                  *
*******************************************************************************}

const
  HDF_BITMAP_ON_RIGHT = $1000;
  HDF_IMAGE = $0800;

  HDI_IMAGE = $0020;
  HDI_DI_SETITEM = $0040;
  HDI_ORDER = $0080;

  HDM_GETITEMRECT = (HDM_FIRST + 7);
  HDM_SETIMAGELIST = (HDM_FIRST + 8);
  HDM_GETIMAGELIST = (HDM_FIRST + 9);
  HDM_ORDERTOINDEX = (HDM_FIRST + 15);
  HDM_CREATEDRAGIMAGE = (HDM_FIRST + 16);
  HDM_GETORDERARRAY = (HDM_FIRST + 17);
  HDM_SETORDERARRAY = (HDM_FIRST + 18);
  HDM_SETHOTDIVIDER = (HDM_FIRST + 19);

  HDN_GETDISPINFOA = (HDN_FIRST-9);
  HDN_GETDISPINFOW = (HDN_FIRST-29);
  HDN_GETDISPINFO = HDN_GETDISPINFOA;
  HDN_BEGINDRAG = (HDN_FIRST-10);
  HDN_ENDDRAG = (HDN_FIRST-11);

type
  PHDItemExA = ^THDItemExA;
  PHDItemExW = ^THDItemExW;
  PHDItemEx = PHDItemExA;

  THDItemExA = packed record
    Mask: Cardinal;
    cxy: Integer;
    pszText: PAnsiChar;
    hbm: HBITMAP;
    cchTextMax: Integer;
    fmt: Integer;
    lParam: LPARAM;
    iImage: Integer;
    iOrder: Integer;
  end;
  THDItemExW = packed record
    Mask: Cardinal;
    cxy: Integer;
    pszText: PWideChar;
    hbm: HBITMAP;
    cchTextMax: Integer;
    fmt: Integer;
    lParam: LPARAM;
    iImage: Integer;
    iOrder: Integer;
  end;
  THDItemEx = THDItemExA;

  PHDNotifyExA = ^THDNotifyExA;
  PHDNotifyExW = ^THDNotifyExW;
  PHDNotifyEx = PHDNotifyExA;
  THDNotifyExA = packed record
    Hdr: TNMHDR;
    Item: Integer;
    Button: Integer;
    PItem: PHDItemExA;
  end;
  THDNotifyExW = packed record
    Hdr: TNMHDR;
    Item: Integer;
    Button: Integer;
    PItem: PHDItemExW;
  end;
  THDNotifyEx = THDNotifyExA;

  PHDDispInfoExA = ^THDDispInfoExA;
  PHDDispInfoExW = ^THDDispInfoExW;
  PHDDispInfoEx = PHDDispInfoExA;

  THDDispInfoExW = packed record
    hdr: TNMHDR;
    Item: Integer;
    mask: Word;
    pszText: PWideChar;
    cchTextMax: Integer;
    iImage: Integer;
    lParam: LPARAM;
  end;

  THDDispInfoExA = packed record
    hdr: TNMHDR;
    Item: Integer;
    mask: Word;
    pszText: PChar;
    cchTextMax: Integer;
    iImage: Integer;
    lParam: LPARAM;
  end;
  THDDispInfoEx = THDDispInfoExA;

  function Header_InsertItemEx(Header: HWnd; Index: Integer;
    const Item: THDItemEx): Integer;
  function Header_GetItemEx(Header: HWnd; Index: Integer;
    var Item: THDItemEx): Bool;
  function Header_SetItemEx(Header: HWnd; Index: Integer; const Item: THDItemEx): Bool;
  function Header_GetItemRect(Header: HWnd; Index: Integer; var Rect: TRect): BOOL;
       // (BOOL)SNDMSG((hwnd), HDM_GETITEMRECT, (WPARAM)iItem, (LPARAM)lprc)

  function Header_SetImageList(Header: HWnd; iImageList: Integer): HImageList;
       // (HIMAGELIST)SNDMSG((hwnd), HDM_SETIMAGELIST, 0, (LPARAM)himl)

  function Header_GetImageList(Header: HWnd): HImageList;
       // (HIMAGELIST)SNDMSG((hwnd), HDM_GETIMAGELIST, 0, 0)

  function Header_OrderToIndex(Header: HWnd; Order: Integer): Integer;
       // (int)SNDMSG((hwnd), HDM_ORDERTOINDEX, (WPARAM)i, 0)

  function Header_CreateDragImage(Header: HWnd; Index: Integer): HImageList;
       // (HIMAGELIST)SNDMSG((hwnd), HDM_CREATEDRAGIMAGE, (WPARAM)i, 0)

  function Header_GetOrderArray(Header: HWnd; iCount: Integer; pi: PInteger): BOOL;
       // (BOOL)SNDMSG((hwnd), HDM_GETORDERARRAY, (WPARAM)iCount, (LPARAM)lpi)

  function Header_SetOrderArray(Header: HWnd; iCount: Integer; pi: PInteger): BOOL;
       // (BOOL)SNDMSG((hwnd), HDM_SETORDERARRAY, (WPARAM)iCount, (LPARAM)lpi)
// lparam = int array of size HDM_GETITEMCOUNT
// the array specifies the order that all items should be displayed.
// e.g.  { 2, 0, 1}
// says the index 2 item should be shown in the 0ths position
//      index 0 should be shown in the 1st position
//      index 1 should be shown in the 2nd position

   function Header_SetHotDivider(Header: HWnd; fPos: BOOL; dw: Longint): Integer;
       // (int)SNDMSG((hwnd), HDM_SETHOTDIVIDER, (WPARAM)fPos, (LPARAM)dw)
// convenience message for external dragdrop
// wParam = BOOL  specifying whether the lParam is a dwPos of the cursor
//              position or the index of which divider to hotlight
// lParam = depends on wParam  (-1 and wParm = FALSE turns off hotlight)

{*******************************************************************************
* End of Extended Header definitions from commctrl.h                           *
*******************************************************************************}

type
  TIntArray = array[0..(MaxInt div SizeOf(Integer)-1)] of Integer;
  PIntArray = ^TIntArray;

  TExHeaderControl = class;

  THeaderSectionStyle = (hsNormal, hsOwnerDraw, hsCallBack);

  TExHeaderStyle = (hsFullDrag, hsDragDrop, hsColumnClick, hsHotTrack);
  TExHeaderStyles = set of TExHeaderStyle;

  TExHeaderSection = class(TCollectionItem)
  private
    FImageIndex: Integer;
    FBitmap: TBitmap;
    FText: string;
    FWidth: Integer;
    FMinWidth: Integer;
    FMaxWidth: Integer;
    FOrder: Integer;
    FStyle: THeaderSectionStyle;
    FBitmapOnRight: Boolean;
    FAlignment: TAlignment;
    FAllowClick: Boolean;
    function GetLeft: Integer;
    function GetRight: Integer;
    procedure SetAlignment(Value: TAlignment);
    procedure SetBitmap(Value: TBitmap);
    procedure SetImageIndex(Value: Integer);
    procedure SetMaxWidth(Value: Integer);
    procedure SetMinWidth(Value: Integer);
    procedure SetStyle(Value: THeaderSectionStyle);
    procedure SetText(const Value: string);
    procedure SetWidth(Value: Integer);
    procedure SetOrder(Value: Integer);
    procedure SetBitmapOnRight(Value: Boolean);
  protected
    function GetDisplayName: string; override;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    property Left: Integer read GetLeft;
    property Right: Integer read GetRight;
  published
    property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
    property AllowClick: Boolean read FAllowClick write FAllowClick default True;
    property Bitmap: TBitmap read FBitmap write SetBitmap default nil;
    property BitmapOnRight: Boolean read FBitmapOnRight write SetBitmapOnRight default False;
    property ImageIndex: Integer read FImageIndex write SetImageIndex default -1;
    property MaxWidth: Integer read FMaxWidth write SetMaxWidth default 10000;
    property MinWidth: Integer read FMinWidth write SetMinWidth default 0;
    property Order: Integer read FOrder write SetOrder default 0;
    property Style: THeaderSectionStyle read FStyle write SetStyle default hsNormal;
    property Text: string read FText write SetText;
    property Width: Integer read FWidth write SetWidth;
  end;

  TExHeaderSections = class(TCollection)
  private
    FExHeaderControl: TExHeaderControl;
    function GetItem(Index: Integer): TExHeaderSection;
    procedure SetItem(Index: Integer; Value: TExHeaderSection);
  protected
    function GetOwner: TPersistent; override;
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(ExHeaderControl: TExHeaderControl);
    function Add: TExHeaderSection;
    property Items[Index: Integer]: TExHeaderSection read GetItem write SetItem; default;
  end;

  TDrawSectionEvent = procedure(ExHeaderControl: TExHeaderControl;
    Section: TExHeaderSection; const Rect: TRect; Pressed: Boolean) of object;
  TSectionNotifyEvent = procedure(ExHeaderControl: TExHeaderControl;
    Section: TExHeaderSection) of object;
  TSectionTrackEvent = procedure(ExHeaderControl: TExHeaderControl;
    Section: TExHeaderSection; Width: Integer;
    State: TSectionTrackState) of object;

  TSectionDragBeginEvent = procedure(ExHeaderControl: TExHeaderControl;
    Section: TExHeaderSection; var AllowDrag: Boolean) of object;
  TSectionDragEndQueryEvent = procedure(ExHeaderControl: TExHeaderControl;
    Section: TExHeaderSection; NewOrderPos: Integer; var AllowDrag: Boolean) of object;

  TSectionCallBackEvent = procedure(ExHeaderControl: TExHeaderControl;
    Section: TExHeaderSection; var ImageIndex: Integer; var Text: String) of object;

  TExHeaderControl = class(TWinControl)
  private
    FImages: TImageList;
    FImageChangeLink: TChangeLink;
    FSections: TExHeaderSections;
    FCanvas: TCanvas;
    FExHeaderStyles: TExHeaderStyles;
    FOnDrawSection: TDrawSectionEvent;
    FOnResize: TNotifyEvent;
    FOnSectionClick: TSectionNotifyEvent;
    FOnSectionDblClick: TSectionNotifyEvent;
    FOnSectionResize: TSectionNotifyEvent;
    FOnDividerDblClick: TSectionNotifyEvent;
    FOnSectionDragBegin: TSectionDragBeginEvent;
    FOnSectionDragEndQuery: TSectionDragEndQueryEvent;
    FOnSectionDragEnd: TSectionNotifyEvent;
    FOnSectionTrack: TSectionTrackEvent;
    FOnSectionCallBack: TSectionCallBackEvent;
    procedure ImageListChange (Sender: TObject);
    procedure SetImages (Value: TImageList);
    procedure SetSections(Value: TExHeaderSections);
    procedure SetExHeaderStyles(Value: TExHeaderStyles);
    procedure UpdateItem(Message, Index: Integer);
    procedure UpdateSection(Index: Integer);
    procedure UpdateSections;
    procedure CNDrawItem(var Message: TWMDrawItem); message CN_DRAWITEM;
    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMWindowPosChanged(var Message: TWMWindowPosChanged); message WM_WINDOWPOSCHANGED;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure DrawSection(Section: TExHeaderSection; const Rect: TRect;
      Pressed: Boolean); dynamic;
    procedure Resize; dynamic;
    procedure SectionClick(Section: TExHeaderSection); dynamic;
    procedure SectionDblClick(Section: TExHeaderSection); dynamic;
    procedure SectionResize(Section: TExHeaderSection); dynamic;
    procedure DividerDblClick(Section: TExHeaderSection); dynamic;
    procedure SectionDragBegin(Section: TExHeaderSection; var AllowOperation: Boolean); dynamic;
    procedure SectionDragEndQuery(Section: TExHeaderSection; NewOrderPos: Integer; var AllowOperation: Boolean); dynamic;
    procedure SectionDragEnd(Section: TExHeaderSection); dynamic;
    procedure SectionTrack(Section: TExHeaderSection; Width: Integer;
      State: TSectionTrackState); dynamic;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SaveSettings(RegKey: String);
    procedure LoadSettings(RegKey: String);
    procedure StoreArray(regKey, colType: String; ColCount: integer;
       const IntArray: array of integer);
    procedure ReadArray(RegKey, ColType: String; ColCount: integer;
       var IntArray: array of integer);
    property Canvas: TCanvas read FCanvas;
  published
    property Align default alTop;
    property DragCursor;
    property DragMode;
    property Enabled;
    property ExtendedStyles: TExHeaderStyles read FExHeaderStyles write SetExHeaderStyles default [];
    property Font;
    property Images: TImageList read FImages write SetImages;
    property Sections: TExHeaderSections read FSections write SetSections;
    property ShowHint;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property Visible;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnDrawSection: TDrawSectionEvent read FOnDrawSection write FOnDrawSection;
    property OnResize: TNotifyEvent read FOnResize write FOnResize;
    property OnSectionCallBack: TSectionCallBackEvent read FOnSectionCallBack write FOnSectionCallBack;
    property OnSectionClick: TSectionNotifyEvent read FOnSectionClick write FOnSectionClick;
    property OnSectionDblClick: TSectionNotifyEvent read FOnSectionDblClick write FOnSectionDblClick;
    property OnDividerDblClick: TSectionNotifyEvent read FOnDividerDblClick write FOnDividerDblClick;
    property OnSectionResize: TSectionNotifyEvent read FOnSectionResize write FOnSectionResize;
    property OnSectionDragBegin: TSectionDragBeginEvent read FOnSectionDragBegin write FOnSectionDragBegin;
    property OnSectionDragEndQuery: TSectionDragEndQueryEvent read FOnSectionDragEndQuery write FOnSectionDragEndQuery;
    property OnSectionDragEnd: TSectionNotifyEvent read FOnSectionDragEnd write FOnSectionDragEnd;
    property OnSectionTrack: TSectionTrackEvent read FOnSectionTrack write FOnSectionTrack;
    property OnStartDrag;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Extended Common Controls', [TExHeaderControl]);
end;

{ TExHeaderSection }

constructor TExHeaderSection.Create(Collection: TCollection);
begin
  FWidth := 50;
  FMaxWidth := 10000;
  FAllowClick := True;
  FBitmap := TBitmap.Create;
  FImageIndex := -1;
  Order := Collection.Count;
  inherited Create(Collection);
end;

destructor TExHeaderSection.Destroy;
var
  I: Integer;
  ExHeaderSection: TExHeaderSection;
begin
  FBitmap.Free;
  if Assigned(Collection) then
  begin
     for I := 0 to Collection.Count-1 do
     begin
       ExHeaderSection := TExHeaderSections(Collection)[I];
       if ExHeadersection.FOrder > Order then
          ExHeaderSection.FOrder := ExHeaderSection.FOrder - 1;
     end;
  end;

  Inherited Destroy;
end;

procedure TExHeaderSection.Assign(Source: TPersistent);
begin
  if Source is TExHeaderSection then
  begin
    Text := TExHeaderSection(Source).Text;
    Width := TExHeaderSection(Source).Width;
    MinWidth := TExHeaderSection(Source).MinWidth;
    MaxWidth := TExHeaderSection(Source).MaxWidth;
    Alignment := TExHeaderSection(Source).Alignment;
    Order := TExHeaderSection(Source).Order;
    BitmapOnRight := TExHeaderSection(Source).BitmapOnRight;
    ImageIndex := TExHeaderSection(Source).ImageIndex;
    Style := TExHeaderSection(Source).Style;
    AllowClick := TExHeaderSection(Source).AllowClick;
    Bitmap.Assign(TExHeaderSection(Source).Bitmap);
    Exit;
  end;
  inherited Assign(Source);
end;

function TExHeaderSection.GetDisplayName: string;
begin
  Result := Text;
  if Result = '' then Result := inherited GetDisplayName;
end;

function TExHeaderSection.GetLeft: Integer;
var
  I: Integer;
begin
  Result := 0;

  if Assigned(Collection) then
  begin
     for I := 0 to Collection.Count-1 do
     begin
       if TExHeaderSections(Collection)[I].Order < Order then
          Inc(Result, TExHeaderSections(Collection)[I].Width);
     end;
  end
  else Result := 0;
end;

function TExHeaderSection.GetRight: Integer;
begin
  Result := Left + Width;
end;

procedure TExHeaderSection.SetAlignment(Value: TAlignment);
begin
  if FAlignment <> Value then
  begin
    FAlignment := Value;
    Changed(False);
  end;
end;

procedure TExHeaderSection.SetBitmap(Value: TBitmap);
begin
   if FBitmap <> Value then
   begin
      FBitmap.Assign(Value);
      Changed(False);
   end;
end;

procedure TExHeaderSection.SetBitmapOnRight(Value: Boolean);
begin
  if FBitmapOnRight <> Value then
  begin
    FBitmapOnRight := Value;
    Changed(False);
  end;
end;

procedure TExHeaderSection.SetImageIndex(Value: Integer);
begin
   if FImageIndex <> Value then
   begin
      FImageIndex := Value;
      Changed(False);
   end;
end;

procedure TExHeaderSection.SetOrder(Value: Integer);
var
   j: Integer;
   ExHeaderSection: TExHeaderSection;
begin
   if (FOrder <> Value) and (Value >= 0) then
   begin
      if Assigned(Collection) then
      begin
         if Value > Collection.Count-1 then Exit;

         for j := 0 to TExHeaderSections(Collection).Count-1  do
         begin
            ExHeaderSection := TExHeaderSections(Collection)[j];

            if Value > FOrder then
            begin
               if (ExHeaderSection.FOrder > FOrder) and
                  (ExHeaderSection.FOrder <= Value) then
                 ExHeaderSection.FOrder := ExHeaderSection.FOrder-1;
            end
            else
            begin
               if (ExHeaderSection.FOrder >= Value) and
                  (ExHeaderSection.FOrder < FOrder) then
                 ExHeaderSection.FOrder := ExHeaderSection.FOrder+1;
            end;
         end;
      end;

      FOrder := Value;
      Changed(False);
   end;
end;

procedure TExHeaderSection.SetMaxWidth(Value: Integer);
begin
  if Value < FMinWidth then Value := FMinWidth;
  if Value > 10000 then Value := 10000;
  FMaxWidth := Value;
  SetWidth(FWidth);
end;

procedure TExHeaderSection.SetMinWidth(Value: Integer);
begin
  if Value < 0 then Value := 0;
  if Value > FMaxWidth then Value := FMaxWidth;
  FMinWidth := Value;
  SetWidth(FWidth);
end;

procedure TExHeaderSection.SetStyle(Value: THeaderSectionStyle);
begin
  if FStyle <> Value then
  begin
    FStyle := Value;
    Changed(False);
  end;
end;

procedure TExHeaderSection.SetText(const Value: string);
begin
  if FText <> Value then
  begin
    FText := Value;
    Changed(False);
  end;
end;

procedure TExHeaderSection.SetWidth(Value: Integer);
begin
  if Value < FMinWidth then Value := FMinWidth;
  if Value > FMaxWidth then Value := FMaxWidth;
  if FWidth <> Value then
  begin
    FWidth := Value;
    Changed(Index < Collection.Count - 1);
  end;
end;

{ TExHeaderSections }

constructor TExHeaderSections.Create(ExHeaderControl: TExHeaderControl);
begin
  inherited Create(TExHeaderSection);
  FExHeaderControl := ExHeaderControl;
end;

function TExHeaderSections.Add: TExHeaderSection;
begin
  Result := TExHeaderSection(inherited Add);
end;

function TExHeaderSections.GetItem(Index: Integer): TExHeaderSection;
begin
  Result := TExHeaderSection(inherited GetItem(Index));
end;

function TExHeaderSections.GetOwner: TPersistent;
begin
  Result := FExHeaderControl;
end;

procedure TExHeaderSections.SetItem(Index: Integer; Value: TExHeaderSection);
begin
  inherited SetItem(Index, Value);
end;

procedure TExHeaderSections.Update(Item: TCollectionItem);
begin
  if Item <> nil then
    FExHeaderControl.UpdateSection(Item.Index) else
    FExHeaderControl.UpdateSections;
end;

{ TExHeaderControl }

constructor TExHeaderControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := [csDoubleClicks];
  Align := alTop;
  Height := 17;
  FSections := TExHeaderSections.Create(Self);
  FCanvas := TControlCanvas.Create;
  TControlCanvas(FCanvas).Control := Self;
  FImageChangeLink := TChangeLink.Create;
  FImageChangeLink.OnChange := ImageListChange;
  ExtendedStyles := [hsFullDrag, hsColumnClick];
end;

destructor TExHeaderControl.Destroy;
begin
  FCanvas.Free;
  FSections.Free;
  inherited Destroy;
end;

procedure TExHeaderControl.CreateParams(var Params: TCreateParams);
begin
  InitCommonControl(ICC_LISTVIEW_CLASSES);
  inherited CreateParams(Params);
  CreateSubClass(Params, 'SysHeader32');
  with Params do
  begin
    if hsColumnClick in FExHeaderStyles then Style := Style or HDS_BUTTONS;
    if hsFullDrag in FExHeaderStyles then Style := Style or (HDS_FULLDRAG);
    if hsDragDrop in FExHeaderStyles then Style := Style or (HDS_DRAGDROP);
    if hsHotTrack in FExHeaderStyles then Style := Style or HDS_HOTTRACK;

    WindowClass.style := WindowClass.style and not (CS_HREDRAW or CS_VREDRAW);
  end;
end;

procedure TExHeaderControl.CreateWnd;
begin
  inherited CreateWnd;
  if (Images <> nil) and Images.HandleAllocated then
     Header_SetImageList(Handle, Images.Handle);
  UpdateSections;
end;

procedure TExHeaderControl.DrawSection(Section: TExHeaderSection;
  const Rect: TRect; Pressed: Boolean);
begin
  if Assigned(FOnDrawSection) then
    FOnDrawSection(Self, Section, Rect, Pressed) else
    FCanvas.FillRect(Rect);
end;

procedure TExHeaderControl.Resize;
begin
  if Assigned(FOnResize) then FOnResize(Self);
end;

procedure TExHeaderControl.SectionClick(Section: TExHeaderSection);
begin
  if Assigned(FOnSectionClick) then FOnSectionClick(Self, Section);
end;

procedure TExHeaderControl.SectionDblClick(Section: TExHeaderSection);
begin
  if Assigned(FOnSectionDblClick) then FOnSectionDblClick(Self, Section);
end;

procedure TExHeaderControl.DividerDblClick(Section: TExHeaderSection);
begin
  if Assigned(FOnDividerDblClick) then FOnDividerDblClick(Self, Section);
end;

procedure TExHeaderControl.SectionResize(Section: TExHeaderSection);
begin
  if Assigned(FOnSectionResize) then FOnSectionResize(Self, Section);
end;

procedure TExHeaderControl.SectionDragbegin(Section: TExHeaderSection; var AllowOperation: Boolean);
begin
  if Assigned(FOnSectionDragBegin) then FOnSectionDragbegin(Self, Section, AllowOperation);
end;

procedure TExHeaderControl.SectionDragEndQuery(Section: TExHeaderSection; newOrderPos: Integer; var AllowOperation: Boolean);
begin
  if Assigned(FOnSectionDragEndQuery) then FOnSectionDragEndQuery(Self, Section, NewOrderPos, AllowOperation);
end;

procedure TExHeaderControl.SectionDragEnd(Section: TExHeaderSection);
begin
  if Assigned(FOnSectionDragEnd) then FOnSectionDragEnd(Self, Section);
end;

procedure TExHeaderControl.SectionTrack(Section: TExHeaderSection;
  Width: Integer; State: TSectionTrackState);
begin
  if Assigned(FOnSectionTrack) then FOnSectionTrack(Self, Section, Width, State);
end;

procedure TExHeaderControl.SetExHeaderStyles(Value: TExHeaderStyles);
begin
   if Value <> FExHeaderStyles then
   begin
      FExHeaderStyles := Value;
      ReCreateWnd;
   end;
end;

procedure TExHeaderControl.SetSections(Value: TExHeaderSections);
begin
  FSections.Assign(Value);
end;

procedure TExHeaderControl.UpdateItem(Message, Index: Integer);
type
  TShowOption = (soText, soBitmap);
  TShowOptions = set of TShowOption;
var
  Item: THDItemEx;
  ShowOptions: TShowOptions;
begin
  with Sections[Index] do
  begin
    FillChar(Item, SizeOf(Item), 0);
    Item.mask := HDI_WIDTH or HDI_FORMAT;
    Item.cxy := Width;

    if Style = hsOwnerDraw then Item.fmt := Item.fmt or HDF_OWNERDRAW
    else
    begin
       ShowOptions := [];

       if Length(Text) > 0 then ShowOptions := ShowOptions + [soText];
       if not Bitmap.Empty then ShowOptions := ShowOptions + [soBitmap];

       case Alignment of
         taLeftJustify: Item.fmt := HDF_LEFT;
         taRightJustify: Item.fmt := HDF_RIGHT;
       else
         Item.fmt := HDF_CENTER;
       end;
       if (FOrder > -1) then
       begin
          Item.Mask := Item.Mask or (HDI_ORDER);
          Item.iOrder := FOrder;
       end;
       if soBitmap in ShowOptions then
       begin
          Item.Mask := Item.Mask or (HDI_BITMAP);
          Item.hbm := Bitmap.handle;
          Item.fmt := Item.fmt or HDF_BITMAP;
       end;

       if (soBitmap in ShowOptions) and
          (soText in ShowOptions) and
          (fBitmapOnRight) then
             Item.fmt := Item.fmt + HDF_BITMAP_ON_RIGHT;

       if Style = hsCallback then
       begin
          Item.Mask := Item.mask or HDI_TEXT or HDI_IMAGE;
          Item.fmt := Item.fmt or (HDF_IMAGE or HDF_STRING);
          item.pszText := LPSTR_TEXTCALLBACK;
          item.iImage := I_IMAGECALLBACK;
       end
       else
       begin
          if soText in ShowOptions then
          begin
             Item.pszText := PChar(Text);
             Item.cchTextMax := Length(Text);
             Item.Mask := Item.mask or HDI_TEXT;
             Item.fmt := Item.fmt or HDF_STRING;
          end;
          if Assigned(FImages) and (ImageIndex > -1) then
          begin
             Item.Mask := Item.Mask or (HDI_IMAGE);
             Item.fmt := Item.fmt or (HDF_IMAGE);
             Item.iImage := ImageIndex;
          end;
       end;
    end;

    SendMessage(Handle, Message, Index, Integer(@Item));
  end;
end;

procedure TExHeaderControl.UpdateSection(Index: Integer);
begin
  if HandleAllocated then UpdateItem(HDM_SETITEM, Index);
end;

procedure TExHeaderControl.UpdateSections;
var
  I: Integer;
begin
  if HandleAllocated then
  begin
   for I := 0 to Header_GetItemCount(Handle) - 1 do
      Header_DeleteItem(Handle, 0);
    for I := 0 to Sections.Count - 1 do UpdateItem(HDM_INSERTITEM, I);
  end;
end;

procedure TExHeaderControl.CNDrawItem(var Message: TWMDrawItem);
var
  SaveIndex: Integer;
begin
  with Message.DrawItemStruct^ do
  begin
    SaveIndex := SaveDC(hDC);
    FCanvas.Handle := hDC;
    FCanvas.Font := Font;
    FCanvas.Brush.Color := clBtnFace;
    FCanvas.Brush.Style := bsSolid;
    DrawSection(Sections[itemID], rcItem, itemState and ODS_SELECTED <> 0);
    FCanvas.Handle := 0;
    RestoreDC(hDC, SaveIndex);
  end;
  Message.Result := 1;
end;

procedure TExHeaderControl.CNNotify(var Message: TWMNotify);
var
  Section: TExHeaderSection;
  TrackState: TSectionTrackState;
  AllowOperation: Boolean;
  myImageIndex: Integer;
  myText: String;
begin
  with PHDNotifyEx(Message.NMHdr)^ do
  begin
    case Hdr.code of
      HDN_ITEMCLICK:
        SectionClick(Sections[Item]);
      HDN_ITEMDBLCLICK:
        SectionDblClick(Sections[Item]);
      HDN_ITEMCHANGED:
        if PItem^.mask and HDI_WIDTH <> 0 then
        begin
          Section := Sections[Item];
          if Section.FWidth <> PItem^.cxy then
          begin
            Section.FWidth := PItem^.cxy;
            SectionResize(Section);
          end;
        end;
      HDN_DIVIDERDBLCLICK:
        DividerDblClick(Sections[Item]);
      HDN_ENDDRAG:
        if PItem^.mask and HDI_ORDER <> 0 then
        begin
          Section := Sections[Item];
          if Section.FOrder <> PItem^.iOrder then
          begin
            AllowOperation := True;
            SectionDragEndQuery(Section, PItem^.iOrder, AllowOperation);
            if AllowOperation then
            begin
               Section.Order := PItem^.iOrder;
               Message.Result := 0;
               SectionDragEnd(Section);
            end
            else Message.Result := 1;
          end;
        end;
      HDN_BEGINDRAG:
        begin
          if Item = -1 then Exit;
          Section := Sections[Item];
          AllowOperation := True;
          SectionDragbegin(Section, AllowOperation);
          if AllowOperation then
             Message.Result := 0
          else Message.Result := 1;
        end;
      HDN_BEGINTRACK, HDN_TRACK, HDN_ENDTRACK:
        begin
          Section := Sections[Item];
          case Hdr.code of
            HDN_BEGINTRACK: begin
               TrackState := tsTrackBegin;
               if Section.MaxWidth = Section.MinWidth then Message.Result := 1
               else message.Result := 0;
            end;
            HDN_ENDTRACK: TrackState := tsTrackEnd;
          else
            TrackState := tsTrackMove;
          end;
          with PItem^ do
          begin
            if cxy < Section.FMinWidth then cxy := Section.FMinWidth;
            if cxy > Section.FMaxWidth then cxy := Section.FMaxWidth;
            SectionTrack(Sections[Item], cxy, TrackState);
          end;
        end;
    end;
  end;
{  with PHDDispInfoEx(Message.NMHdr)^ do
  begin
    case Hdr.code of
      HDN_GETDISPINFO:
        begin
           if Assigned(FOnSectionCallBack) then
           begin
//              myImageIndex := iImage;
//              myText := pszText;
//              FOnSectionCallBack(Self,Sections[Item], myImageIndex, myText);
              iImage := myImageIndex;
              pszText := PChar('Penguin!');
              cchTextMax := Length(myText);
           end;
        end;
    end;
  end;}
end;

procedure TExHeaderControl.WMLButtonDown(var Message: TWMLButtonDown);
var
  Index: Integer;
  Info: THDHitTestInfo;
begin
  Info.Point.X := Message.Pos.X;
  Info.Point.Y := Message.Pos.Y;
  Index := SendMessage(Handle, HDM_HITTEST, 0, Integer(@Info));
  if (Index < 0) or (Info.Flags and HHT_ONHEADER = 0) or
    Sections[Index].AllowClick then inherited;
end;

procedure TExHeaderControl.WMSize(var Message: TWMSize);
begin
  inherited;
  if not (csLoading in ComponentState) then Resize;
end;

procedure TExHeaderControl.WMWindowPosChanged(var Message: TWMWindowPosChanged);
begin
  inherited;
  Invalidate;
end;

procedure TExHeaderControl.ImageListChange(Sender: TObject);
begin
  if HandleAllocated then Header_SetImageList(Handle, TImageList(Sender).Handle);
//    SendMessage(Handle, HDM_SETIMAGELIST, 0, Longint(TImageList(Sender).Handle));
end;

procedure TExHeaderControl.SetImages (Value: TImageList);
begin
  if Assigned (FImages) then
    Images.UnRegisterChanges(FImageChangeLink);

  FImages := Value;

  if Assigned (FImages) then
    begin
      Images.RegisterChanges (FImageChangeLink);
      Header_SetImageList(Handle, Images.Handle);
//      SendMessage(Handle, HDM_SETIMAGELIST, 0, Longint(Images.Handle));
    end
  else
    Header_SetImageList(Handle, 0);
//    SendMessage(Handle, HDM_SETIMAGELIST, 0, Longint(0));

  UpdateSections;
end;

procedure TExHeaderControl.LoadSettings(RegKey: String);
var
  ColCount: integer;
  ColArray: PIntArray;
  x: Integer;
begin
  ColCount := Sections.Count;
  if ColCount > 0 then begin
    GetMem(ColArray, SizeOf(Integer)*ColCount);
    try
      ReadArray(RegKey, 'Sizes', ColCount, ColArray^);
      if ColArray[0] <> -1 then
        for x := 0 to ColCount-1 do
          Sections[x].Width := ColArray[x];
    finally
      FreeMem(ColArray);
    end;
    GetMem(ColArray, SizeOf(Integer)*ColCount);
    try
      ReadArray(RegKey, 'Order', ColCount, ColArray^);
      if ColArray[0] <> -1 then
        for x := 0 to ColCount-1 do
          Sections[x].FOrder := ColArray[x];
      header_SetOrderArray(Handle, ColCount, Pointer(ColArray));
    finally
      FreeMem(ColArray);
    end;
  end;
end;

procedure TExHeaderControl.SaveSettings(RegKey: String);
var
  ColCount: integer;
  ColArray: PIntArray;
  x: Integer;
begin
  ColCount := Sections.Count;
  if ColCount > 0 then begin
    GetMem(ColArray, SizeOf(Integer)*ColCount);
    try
      for x := 0 to ColCount-1 do
        ColArray[x] := Sections[x].Width;
      StoreArray(RegKey, 'Sizes', ColCount, ColArray^);
    finally
      FreeMem(ColArray);
    end;
    GetMem(ColArray, SizeOf(Integer)*ColCount);
    try
      header_GetOrderArray(Handle, ColCount, Pointer(ColArray));
      StoreArray(RegKey, 'Order', ColCount, ColArray^);
    finally
      FreeMem(ColArray);
    end;
  end;
end;

procedure TExHeaderControl.StoreArray(regKey, colType: String; ColCount: integer;
   const IntArray: array of integer);
var
  Reg: TRegIniFile;
  x: integer;
  s: string;
begin
  if ColCount < 1 then exit;
  s := '';
  for x := 0 to ColCount-1 do
    s := s + IntToStr(IntArray[x]) + ',';
  SetLength(s, Length(s)-1);
  Reg := TRegIniFile.Create(RegKey);
  try
    Reg.WriteString('Columns', ColType, s);
  finally
    Reg.Free;
  end;
end;

procedure TExHeaderControl.ReadArray(RegKey, ColType: String; ColCount: integer;
   var IntArray: array of integer);
var
  Reg: TRegIniFile;
  x,y: integer;
  s: string;
begin
  if ColCount < 1 then exit;
  s := '';
  Reg := TRegIniFile.Create(RegKey);
  try
    s := Reg.ReadString('Columns', ColType, '');
  finally
    Reg.Free;
  end;
  if s = '' then begin
    IntArray[0] := -1;
    exit;
  end;
  y := 0;
  for x := 0 to ColCount-1 do begin
    try
      y := Pos(',', s);
      if y = 0 then
        y := Length(s)+1;
      IntArray[x] := StrToInt(Copy(s, 1, y-1));
    except
      IntArray[x] := 0;
    end;
    s := copy(s, y+1, length(s));
    if s = '' then break;
  end;
end;

procedure TExHeaderControl.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = Images then Images := nil;
  end;
end;

{*******************************************************************************
* CommCtrl Functions                                                           *
*******************************************************************************}

function Header_InsertItemEx(Header: HWnd; Index: Integer;
  const Item: THDItemEx): Integer;
begin
  Result := SendMessage(Header, HDM_INSERTITEM, Index, Longint(@Item));
end;

function Header_GetItemEx(Header: HWnd; Index: Integer; var Item: THDItemEx): Bool;
begin
  Result := Bool( SendMessage(Header, HDM_GETITEM, Index, Longint(@Item)) );
end;

function Header_SetItemEx(Header: HWnd; Index: Integer; const Item: THDItemEx): Bool;
begin
  Result := Bool( SendMessage(Header, HDM_SETITEM, Index, Longint(@Item)) );
end;

function Header_GetItemRect(Header: HWnd; Index: Integer; var Rect: TRect): Bool;
begin
  Result:= Bool(SendMessage(Header, HDM_GETITEMRECT, Index, Longint(@Rect)));
end;

function Header_GetImageList(Header: HWnd): HImageList;
begin
  Result:= HImageList(SendMessage(Header, HDM_GETIMAGELIST, 0, 0));
end;

function Header_SetImageList(Header: HWnd; iImageList: Integer): HImageList;
begin
  Result:= HImageList(SendMessage(Header, HDM_SETIMAGELIST, 0, iImageList));
end;

function Header_OrderToIndex(Header: HWnd; Order: Integer): Integer;
begin
  Result:= Integer(SendMessage(Header, HDM_ORDERTOINDEX, Order, 0));
end;

function Header_CreateDragImage(Header: HWnd; Index: Integer): HImageList;
begin
  Result:= HImageList(SendMessage(Header, HDM_CREATEDRAGIMAGE, Index, 0));
end;

function Header_GetOrderArray(Header: HWnd; iCount: Integer; pi: PInteger): BOOL;
begin
  Result:= Bool(SendMessage(Header, HDM_GETORDERARRAY, iCount, Longint(pi)));
end;

function Header_SetOrderArray(Header: HWnd; iCount: Integer; pi: PInteger): BOOL;
begin
  Result:= Bool(SendMessage(Header, HDM_SETORDERARRAY, iCount, Longint(pi)));
end;

function Header_SetHotDivider(Header: HWnd; fPos: BOOL; dw: Longint): Integer;
begin
  Result:= Integer(SendMessage(Header, HDM_SETHOTDIVIDER, Integer(fPos), dw));
end;

{*******************************************************************************
* End CommCtrl Functions                                                       *
*******************************************************************************}

end.
